package net.beaumaris.cachetest.group;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadPoolExecutor;

import net.beaumaris.cachetest.BaseCacheTest;
import net.beaumaris.cachetest.BaseStressingThread;
import net.beaumaris.cachetest.PrintUtils;
import net.beaumaris.cachetest.SharedStats;

import org.apache.log4j.Logger;
import org.infinispan.Cache;
import org.infinispan.distexec.DefaultExecutorService;
import org.infinispan.manager.DefaultCacheManager;
import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

public class GroupingRehashTest extends BaseCacheTest {

   private static final int NUM_KEYS = 8;

   protected final static int NUM_GROUPS = 1;

   private static Logger log = Logger.getLogger(GroupingRehashTest.class.getName());

   @BeforeClass
   public static void init() {
      System.setProperty("java.net.preferIPv4Stack", "true");
   }

   @Before
   public void setup() throws InterruptedException, IOException {
      this.createNodes();
   }

   @After
   public void teardown() throws IOException {

      log.info("Dumping cache contents to file(s)");

      int i = 1;
      for (DefaultCacheManager cm : cacheContainers) {

         Cache<GroupedKey, AllocResource> allocCache = cm.getCache("alloc");
         Cache<String, Session> uuid = cm.getCache("uuid");

         PrintUtils.getCacheContent(allocCache, i);
         PrintUtils.getCacheContent(uuid, i);
         i++;
      }
   }

   @Override
   protected void addNode() throws IOException {

      DefaultCacheManager cm = new DefaultCacheManager("cluster.xml");
      cm.start();
      cacheContainers.add(cm);
   }

   @Test
   public void testRehashJoin() throws Exception {

      final int joinTime = 5000; // ms.

      CountDownLatch latch = new CountDownLatch(NUM_NODES * NUM_THREADS);
      SharedStats stats = new SharedStats();
      ThreadPoolExecutor executor = (ThreadPoolExecutor) Executors.newFixedThreadPool(NUM_NODES * NUM_THREADS);

      // Wait a few seconds for cache to stabilize.  
      Thread.sleep(joinTime);

      this.startThreads(latch, stats, executor);

      // Wait a few seconds before adding nodes.  
      Thread.sleep(joinTime);

      for (int i = 0; i < 3; i++) {

         this.addNode();
         Thread.sleep(joinTime);
      }

      latch.await();
      executor.shutdown();

      log.info(stats);
   }

   @Override
   protected BaseStressingThread createStressingThread(SharedStats stats, DefaultCacheManager cm, CountDownLatch latch) {
      return new GroupStressingThread(stats, cm, latch);
   }

   public class GroupStressingThread extends BaseStressingThread {

      public GroupStressingThread(SharedStats stats, DefaultCacheManager cacheManager, final CountDownLatch latch) {
         super(stats, cacheManager, latch, COUNT);
      }

      @Override
      protected void doCycle() throws Exception {

         // Build a list of keys.
         List<GroupedKey> keys = new ArrayList<GroupedKey>(NUM_KEYS);
         int groupId = r.nextInt(NUM_GROUPS) + 1;
         for (int i = 0; i < NUM_KEYS; i++) {
            int resourceId = r.nextInt(KEY_UPPER_BOUND) + 1;
            keys.add(new GroupedKey(resourceId, groupId));
         }

         Collections.sort(keys, Collections.reverseOrder());

         GroupedKey dummyKey = new GroupedKey(0, groupId);

         AllocateTask allocateTask = new AllocateTask(keys, NUM_NODES * NUM_THREADS);
         Future<String> allocateResult = executor.submit(allocateTask, dummyKey);

         String uuid = allocateResult.get();
         if (uuid == null) {

            stats.allocateFailedCounter.getAndIncrement();
         }

         else {
            stats.allocateSuccessCounter.getAndIncrement();

            sleep(1000);

            DeallocateTask deallocateTask = new DeallocateTask(uuid);

            Future<Integer> deallocateResult = executor.submit(deallocateTask, dummyKey);
            Integer result = deallocateResult.get();

            if (result == 0)
               stats.deallocateSuccessCounter.getAndIncrement();
            else
               stats.deallocateFailedCounter.getAndIncrement();
         }

         // Wait before retrying.
         sleep(1000);
      }

      @Override
      protected void initExecutor(DefaultCacheManager cacheManager) {

         assertEquals(true, cacheManager.getDefaultConfiguration().isGroupsEnabled());
         Cache<GroupedKey, Integer> allocCache = cacheManager.getCache("alloc");
         executor = new DefaultExecutorService(allocCache);
      }
   }
}
